import pathlib
import logging
import json
import hashlib
import numpy as np

TIME_PRECISION = 0.001

labels_file_extension = (".nyxlabel", ".label", ".labels")

"""
Python interface to JSON *.label* files.

It is a poor parent of Julia struct `Dataset` from `PlanarLarvae.jl`.

`Labels` can be used as a `dict`, with (runid, trackid) pairs as keys.
To iterate over its content and access timeseries of labels, you can do:

```
for run_larva in labels:
    _, larva_id = run_larva
    timeseries_data = labels[run_larva]
    ...
```

A `Labels` object will typically describe a single run, but can technically
represent several runs.

The following example function opens such a *.label* file and loads the
associated track data assumed to be available as a *trx.mat* file.
It iterates over the tracks defined in the *.label* file. The same tracks are
expected to be found also in the *trx.mat* file.

```
import numpy as np
from sklearn.metrics import confusion_matrix

def retagged_trxmat_confusion_matrix(label_file):
    # load the .label and associated trx.mat files
    predicted_labels = Labels(label_file)
    labels = predicted_labels.labelspec
    track_data_file = TrxMat(label_file.parent / 'trx.mat')
    expected_labels = track_data_file.read(labels)

    # iterate the tracks
    cm = None
    for run, larva in predicted_labels:

        ypred = np.array([labels.index(label)
                          for label in predicted_labels[(run, larva)].values()])
        yexp = np.stack([expected_labels[label][run][larva]
                         for label in labels])
        iexp, yexp = np.nonzero(yexp.T==1) # the order of dimensions matters
        ypred = ypred[iexp] # if the trx.mat file mentions more tags than the
                            # .label file, some timesteps may appear as untagged

        cm_ = confusion_matrix(yexp, ypred, labels=np.arange(len(labels)))
        cm = cm_ if cm is None else cm + cm_
    return cm
```

The `labelspec` attribute is assumed to be a list, which is valid for *.label*
files generated by automatic tagging.
In contrast, manual tagging stores label names as `Labels.labelspec['names']`,
because label colors are also stored in the `labelspec` attribute.

With taggingbackends==0.9, a related attribute was introduced:
`secondarylabelspec`. To get a unique array of indexable labels, both primary
and secondary labels in a same array, use either `encoding_label_list` or
`decoding_label_list` instead. The two properties typically equal, unless
attribute `decodingspec` is defined.
"""
class Labels:

    def __init__(self, labels=None, labelspec=None, metadata=None, units=None,
            tracking=None):
        # default values (cannot be set in the signature because there are mutable)
        if labels is None: labels = {}
        if labelspec is None: labelspec = {}
        if metadata is None: metadata = {}
        if units is None: units = {'t': 's'}
        if tracking is None: tracking is []
        #
        self.labels, self.metadata = labels, metadata
        self.labelspec, self.units = labelspec, units
        self.secondarylabelspec = None
        self._tracking = tracking
        self._input_labels = None
        self.decodingspec = None
        #
        self.filepath = None # unused attribute; may help to manage data
                             # dependencies in the future
        if isinstance(labels, (str, pathlib.Path)):
            self.filepath = pathlib.Path(labels) if isinstance(labels, str) else labels
            self.load()

    @property
    def tracking(self):
        if not all([isinstance(dep, dict) for dep in self._tracking]):
            self._tracking = [{"filename": dep.name, "sha1": self._checksum(dep)} \
                    for dep in self._tracking]
        return self._tracking

    @tracking.setter
    def tracking(self, files):
        self._tracking = files

    def _checksum(self, dep, alg="sha1"):
        checksum = hashlib.new(alg)
        with open(dep, "rb") as f:
            checksum.update(f.read())
        return checksum.hexdigest()

    @property
    def input_labels(self):
        return self._input_labels

    @input_labels.setter
    def input_labels(self, file):
        self._input_labels = type(self)(file)
        self.metadata = self.input_labels.metadata

    @property
    def camera_framerate(self):
        try:
            return self.metadata["camera"]["framerate"]
        except:
            pass

    @camera_framerate.setter
    def camera_framerate(self, framerate):
        if "camera" not in self.metadata:
            self.metadata["camera"] = {}
        self.metadata["camera"]["framerate"] = framerate

    def __getitem__(self, key):
        if isinstance(key, tuple):
            if len(key) == 1:
                run, = key
                run = self.parse_run_id(run)
                try:
                    return self.labels[run]
                except KeyError:
                    pass
            elif len(key) == 2:
                run, larva = key
                run = self.parse_run_id(run)
                larva = self.parse_larva_id(larva)
                try:
                    return self.labels[run][larva]
                except KeyError:
                    pass
            elif len(key) == 3:
                run, larva, timestamp = key
                run = self.parse_run_id(run)
                larva = self.parse_larva_id(larva)
                timestamp = self.parse_timestamp(timestamp)
                try:
                    return self.labels[run][larva][timestamp]
                except KeyError:
                    pass
            else:
                raise IndexError("too few of many dimensions")
        else:
            run = key
            run = self.parse_run_id(run)
            try:
                return self.labels[run]
            except KeyError:
                pass

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            if len(key) == 1:
                run, = key
                self.set_run(run, value)
            elif len(key) == 2:
                run, larva = key
                self.set_larva(run, larva, value)
            elif len(key) == 3:
                run, larva, timestamp = key
                self.set_timestep(run, larva, timestamp, value)
            elif len(key) < 1:
                raise IndexError("too few dimensions")
            else:
                raise IndexError("too many dimensions")
        else:
            run = key
            self.set_run(run, value)

    def set_run(self, run:str, value:dict):
        run = self.parse_run_id(run)
        self.labels[run] = larvae = {}
        for larva in value:
            self._set_larva(larvae, larva, value[larva])

    def set_larva(self, run:str, larva, value:dict):
        run = self.parse_run_id(run)
        try:
            larvae = self.labels[run]
        except KeyError:
            self.labels[run] = larvae = {}
        self._set_larva(larvae, larva, value)

    def _set_larva(self, larvae, larva, value):
        larva = self.parse_larva_id(larva)
        larvae[larva] = timesteps = {}
        for timestamp in value:
            self._set_timestep(timesteps, timestamp, value[timestamp])

    def set_timestep(self, run:str, larva, timestamp:float, value):
        run = self.parse_run_id(run)
        try:
            larvae = self.labels[run]
        except KeyError:
            self.labels[run] = larvae = {}
        larva = self.parse_larva_id(larva)
        try:
            timesteps = larvae[larva]
        except KeyError:
            larvae[larva] = timesteps = {}
        self._set_timestep(timesteps, timestamp, value)

    def _set_timestep(self, timesteps, timestamp, value):
        timestamp = self.parse_timestamp(timestamp)
        timesteps[timestamp] = value

    def parse_run_id(self, run:str):
        return run

    def parse_larva_id(self, larva):
        return int(larva)

    def parse_timestamp(self, time:float):
        return np.around(time, int(-np.log10(TIME_PRECISION)))

    def __iter__(self):
        for run in self.labels:
            for larva in self.labels[run]:
                yield run, larva

    def dump(self, file, format="json", indent=2, **kwargs):
        if format != "json":
            raise ArgumentError(f"unsupported format: {format}")
        if isinstance(file, str):
            file = pathlib.Path(file)
        if isinstance(file, pathlib.Path):
            if file.is_dir():
                file = file / "labels.json"
            with file.open('w') as fp:
                json.dump(self, fp, cls=LabelEncoder, indent=indent, **kwargs)
        else:
            json.dump(self, file, cls=LabelEncoder, indent=indent, **kwargs)

    def load(self, file=None, format="json"):
        if format != "json":
            raise ArgumentError(f"unsupported format: {format}")
        if file is None:
            file = self.filepath
        elif isinstance(file, str):
            file = pathlib.Path(file)
        if isinstance(file, pathlib.Path):
            if file.is_dir():
                file = file / "labels.json"
            with file.open('r') as fp:
                new_self = json.load(fp, cls=LabelDecoder)
        else:
            new_self = json.load(file, cls=LabelDecoder)
        self.labels, self.metadata = new_self.labels, new_self.metadata
        self.labelspec, self.units = new_self.labelspec, new_self.units
        self.secondarylabelspec = new_self.secondarylabelspec
        # no `decodingspec` in label files
        self.tracking = new_self.tracking
        return self

    def to_hierarchical_format(self):
        assert len(self.labels) == 1
        run = next(iter(self.labels.keys()))
        data = {}
        metadata = {"id": run}
        if self.metadata:
            metadata.update(self.metadata)
            if "run" in metadata and metadata["run"] == run:
                metadata.pop("run")
        data["metadata"] = metadata
        assert self.units
        data["units"] = self.units
        if self.labelspec:
            data["labels"] = self.labelspec
        if self.secondarylabelspec:
            data["secondarylabels"] = self.secondarylabelspec
        if self.tracking:
            if isinstance(self.tracking, dict):
                data["dependencies"] = self.tracking
            elif len(self.tracking) == 1:
                data["dependencies"] = self.tracking[0]
            else:
                data["dependencies"] = self.tracking
        data["data"] = larvae = []
        for larva, timeseries in self.labels[run].items():
            assert larva not in ("metadata", "units", "labels")
            timestamps, labels = list(timeseries.keys()), list(timeseries.values())
            larvae.append({
                "id": str(larva),
                "t": self.parse_timestamp(timestamps).tolist(),
                "labels": self.encode(labels)})
        return data

    def from_hierarchical_format(self, data):
        self.metadata = dict(data["metadata"])
        run = self.metadata.pop("id")
        self.units = data.get("units", {})
        self.labelspec = data.get("labels", {})
        self.secondarylabelspec = data.get("secondarylabels", [])
        self.decodingspec = None # ensure it is not set
        self._tracking = data.get("dependencies", [])
        if isinstance(self._tracking, dict):
            self._tracking = [self._tracking]
        tracks = data["data"]
        self.labels = {}
        self.labels[run] = larvae = {}
        for track in tracks:
            larva = int(track["id"])
            larvae[larva] = {timestamp: self.decode(label) \
                    for timestamp, label in zip(track["t"], track["labels"])}
        return self

    """
    List of str: all different labels including primary and secondary labels.
    """
    @property
    def encoding_label_list(self):
        if isinstance(self.labelspec, dict):
            labelset = self.labelspec['names']
        else:
            labelset = self.labelspec
        if self.secondarylabelspec:
            labelset = labelset + self.secondarylabelspec
        return labelset

    """
    List of str: all different labels including primary and secondary labels.
    """
    @property
    def decoding_label_list(self):
        labelspec = self.decodingspec
        if labelspec is None:
            return self.encoding_label_list
        if self.secondarylabelspec:
            labelset = labelset + self.secondarylabelspec
        return labelset

    """
    Encode the text labels as indices (`int` or `list` of `int`).

    Labels are 1-indexed. If shifted down, the indices apply to attribute
    `labelspec`.
    """
    def encode(self, label=None):
        if label is None:
            encoded = label = self
            for run_larva in label:
                label[run_larva] = self.encode(label[run_larva])
        elif isinstance(label, dict):
            encoded = {t: self.encode(l) for t, l in label.items()}
        else:
            labelset = self.encoding_label_list
            if isinstance(label, str):
                encoded = labelset.index(label) + 1
            elif isinstance(label, int):
                encoded = label
                logging.debug('label(s) already encoded')
            else:
                encoded = [labelset.index(l) + 1 for l in label]
        return encoded

    """
    Decode the label indices as text (`str` or `list` of `str`).

    Text labels are picked in `labelspec`, or `decodingspec` instead if defined.

    `decodingspec` is set to decode the encoded output from a tagger that
    defines redundant labels, which are remapped onto labels in `labelspec`.
    These may include several sections in their config file, including
    `original_behavior_labels`, `behavior_labels`, `remapped_behavior_labels`.

    Note that `decodingspec` should not be set to decode a label file. A label
    file should use and mention the remapped labels only, if the latter are
    defined by the tagger used to generate the label file. In the case of
    decoding a label file, `labelspec` only should be defined in the `Labels`
    object.

    To forget about these considerations, call `load_model_config` first if
    processing the output of a tagger, or do not call that method otherwise.

    MaggotUBA's `predict_model` does not call `decode` but directly indexes into
    `behavior_labels`. As a consequence, the above subtleties do not apply to
    MaggotUBA.
    """
    def decode(self, label=None):
        if label is None:
            decoded = label = self
            for run_larva in label:
                label[run_larva] = self.decode(label[run_larva])
        elif isinstance(label, dict):
            decoded = {t: self.decode(l) for t, l in label.items()}
        else:
            labelset = self.decoding_label_list
            if isinstance(label, int):
                decoded = labelset[label-1]
            elif isinstance(label, str):
                decoded = label
                logging.debug('label(s) already decoded')
            else:
                decoded = [labelset[l-1] for l in label]
        return decoded

    def load_model_config(self, config):
        try:
            self.labelspec = config["remapped_behavior_labels"]
        except KeyError:
            self.labelspec = config["behavior_labels"]
        else:
            self.decodingspec = config["behavior_labels"]

class LabelEncoder(json.JSONEncoder):
    def default(self, labels):
        if isinstance(labels, Labels):
            return labels.to_hierarchical_format()
        return json.JSONEncoder.default(self, labels)

class LabelDecoder(json.JSONDecoder):
    def decode(self, s):
        data = json.JSONDecoder.decode(self, s)
        return Labels().from_hierarchical_format(data)

